# Function to turn text like "!ENVHAS(x) && !(INPUTHAS(abc) || OUTPUTHAS(def))" into an if() condition
# INPUT: text of the form above
# OUTPUT: text of the form "(('x'%in%VARNAMES) && !((length(grep('abc', INPUT))>0) || (length(grep('def', OUTPUT))>0)))"
#         or an object of class "try-error" (with attribute 'message')
# See feedbackParser.R for full list of possible feedback sections ('validSections')
# Called by rFeedback()
#
makeCondition = function(txt) {
  tryError = list(); class(tryError)="try-error"
  
  # CONDITION()  # includes use of backticks holding named or numbered variables in the environment
  locs = gregexpr("CONDITION[[:space:]]*[(]", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
      pExpr = try(getParen(orig))
      if (is(pExpr, "try-error")) return(pExpr)
      condition = pExpr$inside
      rhLoc = pExpr$rhLoc
      # process backticked variable names/numbers
      bt = gregexpr('`', condition)[[1]]
      if (length(bt)==1 && bt==-1) bt = numeric(0)
      if (length(bt)%%2 != 0) {
        attr(tryError, "message") = "unbalanced backticks"
        return(tryError)
      }
      nBt2 = length(bt)
      if (nBt2>0) {
        btMat = cbind(bt[seq(1,nBt2,2)] + 1, bt[seq(2,nBt2,2)] - 1)
        btStr = str_trim(unique(sapply(1:(nBt2/2), function(i) substring(condition, btMat[i,1], btMat[i,2]))))
        if (any(nchar(btStr)) == 0) {
          attr(tryError, "message") = "cannot have empty backticks"
          return(tryError)
        }
        asNumeric = suppressWarnings(as.numeric(btStr))
        isNumeric = !is.na(asNumeric)
        nBt = length(btStr)
        nBtNumeric = sum(isNumeric)
        # Build pre-condition, checking that backticked variable exist in the environment
        preCondition = character(0)
        if (nBtNumeric > 0) {
          preCondition = paste0('length(VARNAMES)>=', max(asNumeric[isNumeric]))
          if (nBt > nBtNumeric) preCondition = paste0(preCondition, ' && ')
        }
        if (nBt > nBtNumeric) {
          vars = paste0('"', btStr[!isNumeric], '"', collapse=",")
          preCondition = paste0(preCondition, '!any(is.na(match(c(', vars, '), VARNAMES)))')
        }
        preCondition = paste0('(', preCondition, ')')
        # Build main condition by substituting for backticked variables
        for (b in 1:nBt) {
          pattern = paste0('`', btStr[b], '`')
          if (isNumeric[b]) {
            replacement = paste0('VARS[[', asNumeric[b], ']]')
          } else {
            replacement = paste0('VARS[["', btStr[b], '"]]')
          }
          condition = gsub(pattern, replacement, condition)
        }
        # Combine precondition with main condtion
        mid = paste0('(', preCondition, ' && ', condition, ')')
      } else {
        mid = paste0("(", condition, ")")
      }
      # Assemble final code
      if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
      txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
    } # end for each occurance
  } # end if any ENVHASE
  
  
  # ENVHAS()
  locs = gregexpr("ENVHAS[[:space:]]*[(]", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
      pExpr = try(getParen(orig))
      if (is(pExpr, "try-error")) return(pExpr)
      condition = pExpr$inside
      rhLoc = pExpr$rhLoc
      mid = paste0("('", condition, "' %in% VARNAMES)")
      if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
      txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
    } # end for each occurance
  } # end if any ENVHASE
  
  # ENVLACKS()
  locs = gregexpr("ENVLACKS[[:space:]]*[(]", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
      pExpr = try(getParen(orig))
      condition = pExpr$inside
      rhLoc = pExpr$rhLoc
      mid = paste0("(!'", condition, "' %in% VARNAMES)")
      if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
      txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
    } # end for each occurance
  } # end if any ENVLACKS
  
  # ENVEMPTY
  locs = gregexpr("ENVEMPTY", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      Sel = grep("^[[:space:]]*[(]", substring(txt, locs[i]+8))
      if (length(Sel)>0) {
        attr(tryError, "message") = "'ENVEMPTY' should not have a parenthesized condition"
        return(tryError)
      }
      txt = paste0(substring(txt, 1, locs[i] - 1), 
                   "(length(VARNAMES)==0)", 
                   substring(txt, locs[i]+8))
    } # end for each occurance
  } # end if any ENVEMPTY
  
  # ENVNOTEMPTY
  locs = gregexpr("ENVNOTEMPTY", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      Sel = grep("^[[:space:]]*[(]", substring(txt, locs[i]+11))
      if (length(Sel)>0) {
        attr(tryError, "message") = "'ENVNOTEMPTY' should not have a parenthesized condition"
        return(tryError)
      }
      txt = paste0(substring(txt, 1, locs[i] - 1), 
                   "(length(VARNAMES)>0)", 
                   substring(txt, locs[i]+11))
    } # end for each occurance
  } # end if any ENVNOTEMPTY
  
  # VARCLASSIS()
  locs = gregexpr("VARCLASSIS[[:space:]]*[(]", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
      pExpr = try(getParen(orig))
      if (is(pExpr, "try-error")) return(pExpr)
      condition = pExpr$inside
      rhLoc = pExpr$rhLoc
      comma = regexpr(",", condition)[1]
      if (comma == -1) {
        attr(tryError, "message") = "expression for 'VARCLASSIS' should be '(var,className)'"
        return(tryError)
      }
      var = str_trim(substring(condition, 1, comma - 1))
      className = str_trim(substring(condition, comma + 1))
      if (nchar(var)==0) {
        attr(tryError, "message") = "missing 'var' for 'VARCLASSIS(var,className)'"
        return(tryError)
      }
      if (nchar(className)==0) {
        attr(tryError, "message") = "missing 'className' for 'VARCLASSIS(var,className)'"
        return(tryError)
      }
      vNum = suppressWarnings(as.numeric(var))
      if (is.na(vNum)) {
        mid = paste0("('", var, "' %in% VARNAMES && is(VARS[['", var, "']], '",className, "'))")
      } else {
        mid = paste0("(length(VARNAMES)>=", var, " && is(VARS[[", var, "]], '",className, "'))")
      }
      if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
      txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
    } # end for each occurance
  } # end if any VARCLASSIS
  
  # VARCLASSISNT()
  locs = gregexpr("VARCLASSISNT[[:space:]]*[(]", txt)[[1]]
  if (locs[1] != -1) {
    locs = rev(locs)
    for (i in 1:length(locs)) {
      orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
      pExpr = try(getParen(orig))
      if (is(pExpr, "try-error")) return(pExpr)
      condition = pExpr$inside
      rhLoc = pExpr$rhLoc
      comma = regexpr(",", condition)[1]
      if (comma == -1) {
        attr(tryError, "message") = "expression for 'VARCLASSIS' should be '(var,className)'"
        return(tryError)
      }
      var = str_trim(substring(condition, 1, comma - 1))
      className = str_trim(substring(condition, comma + 1))
      if (nchar(var)==0) {
        attr(tryError, "message") = "missing 'var' for 'VARCLASSISNT(var,className)'"
        return(tryError)
      }
      if (nchar(className)==0) {
        attr(tryError, "message") = "missing 'className' for 'VARCLASSISNT(var,className)'"
        return(tryError)
      }
      vNum = suppressWarnings(as.numeric(var))
      if (is.na(vNum)) {
        mid = paste0("('", var, "' %in% VARNAMES && !is(VARS[['", var, "']], '",className, "'))")
      } else {
        mid = paste0("(length(VARNAMES)<", var, " && !is(VARS[[", var, "]], '",className, "'))")
      }
      if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
      txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
    } # end for each occurance
  } # end if VARCLASSISNT

  # INPUT / OUTPUT / ERRORMESSAGE / ERRORCODE HAS()
  for (source in c("INPUT", "OUTPUT", "ERRORMESSAGE", "ERRORCODE")) {
    locs = gregexpr(paste0(source,"HAS[*]?[[:space:]]*[(]"), txt)[[1]]
    if (locs[1] != -1) {
      locs = rev(locs)
      for (i in 1:length(locs)) {
        orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
        pExpr = try(getParen(orig))
        if (is(pExpr, "try-error")) return(pExpr)
        condition = pExpr$inside
        rhLoc = pExpr$rhLoc
        mid = paste0("(length(grep('", condition, "', ", source, ", fixed=", pExpr$hasStar, ")) > 0)")
        if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
        txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
      } # end for each occurance
    } # end if any INPUTHAS / OUTPUTHAS
  } # end for two sources
  
  # INPUT / OUTPUT / ERRORMESSAGE / ERRORCODE LACKS()
  for (source in c("INPUT", "OUTPUT", "ERRORMESSAGE", "ERRORCODE")) {
    locs = gregexpr(paste0(source,"LACKS[*]?[[:space:]]*[(]"), txt)[[1]]
    if (locs[1] != -1) {
      locs = rev(locs)
      for (i in 1:length(locs)) {
        orig = if (i==1) substring(txt, locs[1]) else substring(txt, locs[i], locs[i-1] - 1)
        pExpr = try(getParen(orig))
        if (is(pExpr, "try-error")) return(pExpr)
        condition = pExpr$inside
        rhLoc = pExpr$rhLoc
        mid = paste0("(length(grep('", condition, "', ", source, ", fixed=", pExpr$hasStar, ")) == 0)")
        if (!is.na(rhLoc)) rhs = ifelse(is.na(rhLoc), "", substring(txt, locs[i] + rhLoc - 1))
        txt = paste0(substring(txt, 1, locs[i] - 1), mid, rhs)
      } # end for each occurance
    } # end if any INPUTLACKS / OUTPUTLACKS
  } # end for two sources
  
  p = try(parse(text=txt), silent=TRUE)
  if (is(p, "try-error")) {
    attr(tryError, "message") = paste0('invalid expression: ', txt, '\n(check parentheses, etc.)')
    return(tryError)
  }
  
  return(txt)
} # end makeCondition()
